# Final Frontier
# Civilization 4 (c) 2007 Firaxis Games

# Designed & Programmed by -	Jon 'Trip' Shafer

from CvPythonExtensions import *
from PyHelpers import PyPlayer
import CvUtil
import CvGameUtils
import CvEventInterface
import math
import pickle

from CvSolarSystem import *

# globals
gc = CyGlobalContext()
localText = CyTranslator()

DefaultUnitAI = UnitAITypes.NO_UNITAI

class CvAI:
	
	def __init__(self):
		
		printd("\n\n\n\n\n init-ing CvAI \n\n\n\n\n")
		self.iNumPlayerAIInfos = 0
		self.apPlayerAIInfos = []
		
	def addPlayerAIInfo(self, pPlayerAIInfo):
		self.apPlayerAIInfos.append(pPlayerAIInfo)
		self.changeNumPlayerAIInfos(1)
		printd("self.iNumPlayerAIInfos")
		printd(self.iNumPlayerAIInfos)
		
	def getPlayerAIInfo(self, iID):
		printd("self.iNumPlayerAIInfos = %d" % self.iNumPlayerAIInfos)
		printd("        iID = %d" % iID)
		
		if (iID < 0 or iID > self.iNumPlayerAIInfos):
			fassert
			
		return self.apPlayerAIInfos[iID]
		
	def getNumPlayerAIInfos(self):
		return self.iNumPlayerAIInfos
	def setNumPlayerAIInfos(self, iValue):
		self.iNumPlayerAIInfos = iValue
	def changeNumPlayerAIInfos(self, iChange):
		self.iNumPlayerAIInfos += iChange
	
	def initPlayerAIInfos(self):
		
		printd("\nInitializing Player AI object array within CvAI")
		
		for iPlayerLoop in range(gc.getMAX_CIV_PLAYERS()):
			
			pInfo = CvPlayerAIInfo(iPlayerLoop)
			
			self.addPlayerAIInfo(pInfo)
	
	def doSavePlayerAIInfos(self):
		
		printd("\nStoring player AI Object information in CvPlayer scriptData")
		
		for iPlayerLoop in range(gc.getMAX_CIV_PLAYERS()):
			
			pPlayer = gc.getPlayer(iPlayerLoop)
			pPlayerAIInfo = self.getPlayerAIInfo(iPlayerLoop)
			
			aSaveInfo = pPlayerAIInfo.saveData()
			
			pPlayer.setScriptData(pickle.dumps(aSaveInfo))
			printd("Saving array to player %d" %(iPlayerLoop))
			printd(aSaveInfo)
		
	def doLoadPlayerAIInfos(self):
		
		printd("\nLoading player AI Object information from CvPlayer scriptData and creating new objects")
		
		for iPlayerLoop in range(gc.getMAX_CIV_PLAYERS()):
			
			pPlayer = gc.getPlayer(iPlayerLoop)
			
			# Add loaded player to member array
			pInfo = CvPlayerAIInfo(iPlayerLoop)
			self.addPlayerAIInfo(pInfo)
			
			# Get data to be imported
			aLoadInfo = pickle.loads(pPlayer.getScriptData())
			
			# Transfer load data
			self.getPlayerAIInfo(iPlayerLoop).loadData(aLoadInfo)
			
			printd("Loading array from player %d scriptData" %(iPlayerLoop))
			printd(aLoadInfo)
			
			
			
		
	
##########################################################
##########################################################
##########################################################

#		CITY UPDATE

##########################################################
##########################################################
##########################################################
	
	def doCityAIUpdate(self, pCity, iPopulationToAssign = -1):
		
		FinalFrontier = CvEventInterface.getEventManager()
		
		pSystem = FinalFrontier.getSystemAt(pCity.getX(), pCity.getY())
		
		iOwner = pCity.getOwner()
		
		iMaxSupportablePop = pSystem.getPopulationLimit(true)
		
		if (iPopulationToAssign > iMaxSupportablePop):
			iPopulationToAssign = iMaxSupportablePop
		
		printd("pSystem.getPopulation() = %d" % (pSystem.getPopulation()))
		#printd(pSystem.getPopulation())
		printd("pSystem.getPopulationLimit() = %d" % (pSystem.getPopulationLimit()))
		#printd(pSystem.getPopulationLimit())
		printd("iPopulationToAssign = %d" % (iPopulationToAssign))
		#printd(iPopulationToAssign)
		
		if (iPopulationToAssign == -1):
			iPopulationToAssign = pCity.getPopulation()
		else:
			if (pSystem.getPopulation() + iPopulationToAssign > iMaxSupportablePop):
				iPopulationToAssign = iMaxSupportablePop - pSystem.getPopulation()
		
		if (iPopulationToAssign <= 0):
			return
		
#		printd("\n\nDoing AI for city %s" %(pCity.getName()))
		
		aaiPlanetValues = []
		
		# Loop through all planets and ascertain their value
		for iPlanetLoop in range(pSystem.getNumPlanets()):
			
			pPlanet = pSystem.getPlanetByIndex(iPlanetLoop)
			
			iPlanetRingID = pPlanet.getOrbitRing()
			
			iFood = pPlanet.getTotalYield(iOwner, 0)
			iProd = pPlanet.getTotalYield(iOwner, 1)
			iGold = pPlanet.getTotalYield(iOwner, 2)
			aaiPlanetValues.append([iFood, iProd, iGold, iPlanetRingID])
			
			# Reset the population so it can be redistributed
			if (iPopulationToAssign == pCity.getPopulation()):
				pPlanet.setPopulation(0)
		
		iYieldNeededMost = 0	# Default yield needed most is food
		
		iBestPlanet = 0
			
		FinalFrontier.updatePlotYield(pCity.plot())
		
		iCurrentFood = pCity.getYieldRate(0)
		iCurrentProduction = pCity.getYieldRate(1)
		iSurplusFood = iCurrentFood - (pCity.getPopulation() * gc.getDefineINT("FOOD_CONSUMPTION_PER_POPULATION"))
		
		printd("iPopulationToAssign = %d" %(iPopulationToAssign))
		#printd(iPopulationToAssign)
		
		# Place population 1 by 1
		for iPopLoop in range(iPopulationToAssign):
			
			printd("\n   iPopLoop = %d" % (iPopLoop))
			#printd(iPopLoop)
			
			# Determine what the new most important yield is
			
			# Always need food to grow
			if (iSurplusFood < 3):
				iYieldNeededMost = 0
			else:
				iYieldNeededMost = 1
			
			# Barebones Production
			if (iCurrentProduction == 0 and iSurplusFood >= 2):
				iYieldNeededMost = 1
			
			#If avoid growth is on, prioritize commerce or production -- TC01
			if (pCity.AI_avoidGrowth()):
				iYieldNeededMost = gc.getGame().getSorenRandNum(2, "AI Population Assignment Logic") + 1
				
#			printd("   iYieldNeededMost")
#			printd(iYieldNeededMost)
			
			aaiTempPlanetValues = []
			
			# Create a new list based on the most needed yield
			for aiPlanetValueLoop in aaiPlanetValues:
				iOtherYield = 0
				
				for iLoop in range(3):
					if (iLoop != iYieldNeededMost):
						iOtherYield += aiPlanetValueLoop[iLoop]
				
				aaiTempPlanetValues.append([aiPlanetValueLoop[iYieldNeededMost], iOtherYield, aiPlanetValueLoop[3]])	# The 3 here is the Ring ID
				
			# Sort the list based on value
			aaiTempPlanetValues.sort()
			aaiTempPlanetValues.reverse()
			
			printd("   Sorted list of most valued planets based on yield")
			printd(aaiTempPlanetValues)
			
			for aiTempPlanetValue in aaiTempPlanetValues:
				
				pCurrentPlanet = pSystem.getPlanet(aiTempPlanetValue[2])
				
				if (pCurrentPlanet.getPopulation() < pCurrentPlanet.getPopulationLimit(iOwner)):
					
					iBestPlanet = aiTempPlanetValue[2]
					break
			
			printd("   iBestPlanetRing = %d" % (iBestPlanet))
			#printd(iBestPlanet)
			
			# Not a valid ring, return (no usable planets!)
			if (iBestPlanet == 0):
				break
			
			pBestPlanet = pSystem.getPlanet(iBestPlanet)
			
			# Add the population
			FinalFrontier.doAddPopulationToPlanet(pSystem, iBestPlanet)
			
			iCurrentFood += pBestPlanet.getTotalYield(iOwner, 0)
			iCurrentProduction += pBestPlanet.getTotalYield(iOwner, 1)
			iSurplusFood = iCurrentFood - (pCity.getPopulation() * gc.getDefineINT("FOOD_CONSUMPTION_PER_POPULATION"))
			
#			printd("   iSurplusFood")
#			printd(iSurplusFood)
			
		FinalFrontier.updatePlotYield(pCity.plot())
		
	def doCityAIProduction(self, pCity):
		
		FinalFrontier = CvEventInterface.getEventManager()
		
		bOverride = false
		
		iPlayer = pCity.getOwner()
		pPlayer = gc.getPlayer(iPlayer)
		pCivilization = gc.getCivilizationInfo( pPlayer.getCivilizationType())
		iTeam = pPlayer.getTeam()
		pTeam = gc.getTeam(iTeam)
		pSystem = FinalFrontier.getSystemAt(pCity.getX(), pCity.getY())
		
		printd("\n\nDoing AI for %s's City %s" %(pPlayer.getName(), pCity.getName()))
		
		# Small chance of randomly letting the AI do whatever it wants... sometimes the weights can go overboard :)
		iRand = CyGame().getSorenRandNum(100, "Final Frontier: Random roll to see if City AI override exits")
		if (iRand < 15):
			return false
		
		iHabitationSystem = pCivilization.getCivilizationBuildings(gc.getInfoTypeForString('BUILDINGCLASS_HABITATION_SYSTEM')) # this civ's habitation system
		iExtHabitationSystem = pCivilization.getCivilizationBuildings(gc.getInfoTypeForString('BUILDINGCLASS_EXTENDED_HABITATION_SYSTEM')) # this civ's extended habitation system
		
		iCapitalShipyard = pCivilization.getCivilizationBuildings(gc.getInfoTypeForString('BUILDINGCLASS_CAPITAL_SHIPYARD')) # this civ's CAPITAL_SHIPYARD
		iTrainingCompound = pCivilization.getCivilizationBuildings(gc.getInfoTypeForString('BUILDINGCLASS_TRAINING_COMPOUND')) # this civ's TRAINING_COMPOUND
		
		iSportsArena = pCivilization.getCivilizationBuildings(gc.getInfoTypeForString('BUILDINGCLASS_SPORTS_ARENA')) # this civ's SPORTS_ARENA
		iRecyclingCenter = pCivilization.getCivilizationBuildings(gc.getInfoTypeForString('BUILDINGCLASS_RECYCLING_CENTER')) # this civ's RECYCLING_CENTER
		
		iNutritionFacility = pCivilization.getCivilizationBuildings(gc.getInfoTypeForString('BUILDINGCLASS_NUTRITION_FACILITY')) # this civ's NUTRITION_FACILITY
		iMiningFacility = pCivilization.getCivilizationBuildings(gc.getInfoTypeForString('BUILDINGCLASS_MINING_FACILITY')) # this civ's MINING_FACILITY
		iNanoExtraction = pCivilization.getCivilizationBuildings(gc.getInfoTypeForString('BUILDINGCLASS_NANOEXTRACTION_UPGRADE')) # this civ's nanoextraction upgrade
		iMagLevNetwork = pCivilization.getCivilizationBuildings(gc.getInfoTypeForString('BUILDINGCLASS_MAG_LEV_NETWORK')) # this civ's MAG_LEV_NETWORK
		iCommercialSatellites = pCivilization.getCivilizationBuildings(gc.getInfoTypeForString('BUILDINGCLASS_COMMERCIAL_SATELLITES')) # this civ's COMMERCIAL_SATELLITES
		
		# Cheat a little bit, need to get more Construction Ships

		iConstructionShipClass = gc.getInfoTypeForString(gc.getDefineSTRING('CONSTRUCT_SHIP'))
		iConstructionShip = pCivilization.getCivilizationUnits( iConstructionShipClass ) # get this civilization's unit for this unitclass
							
		# Size requirement
		if (pPlayer.getNumCities() > 1):
			if (pCity.getPopulation() >= 4):
				
				iNumUnits = 0
				
				pyPlayer = PyPlayer(iPlayer)
				apUnitList = pyPlayer.getUnitList()
				for pUnitLoop in apUnitList:
					if (pUnitLoop.getUnitType() == iConstructionShip):
						iNumUnits += 1
				
				# Either 0 ships or less than certain # of cities...
				iNumCitiesThreshold = (pPlayer.getNumCities() / 2) + 1
				if ((iNumUnits < iNumCitiesThreshold) or (iNumUnits == 0)):
					
					iRand = CyGame().getSorenRandNum(100, "Rolling to see if AI should be forced to build a Construction Ship")
					
					# 15% Chance of forcing AI to build one of these things
					if (iRand < 15):
						
						printd("   Forcing city to make a Construction Ship :)")
						
						pCity.pushOrder(OrderTypes.ORDER_TRAIN, iConstructionShip, -1, False, False, False, True)
						bOverride = true
						
						return bOverride
			
			
		
		#######################################################################
		#######################################################################
		
		# Determine Weighting for what the AI should build
		
		#######################################################################
		#######################################################################
		
		printd("   Determining weights for various things to construct")
		
		# If any one weight is less than this value then we will let the C++ AI do whatever it wants
		iWeightNeeded = 100
		iDefaultWeight = 0
		
		aszWeightList = []
		
		iIterator = 0
		
		iMilitaryWeightType = iIterator
		aszWeightList.append("MilitaryWeightType")
		iIterator += 1
		iPopulationWeightType = iIterator
		aszWeightList.append("PopulationWeightType")
		iIterator += 1
		iProductionWeightType = iIterator
		aszWeightList.append("ProductionWeightType")
		iIterator += 1
		iCommerceWeightType = iIterator
		aszWeightList.append("CommerceWeightType")
		iIterator += 1
		iHappyWeightType = iIterator
		aszWeightList.append("HappyWeightType")
		iIterator += 1
		iHealthWeightType = iIterator
		aszWeightList.append("HealthWeightType")
		iIterator += 1
		iFoodWeightType = iIterator
		aszWeightList.append("FoodWeightType")
		iIterator += 1
		
		iNumWeights = iIterator
		
		aiWeights = []
		for iWeightLoop in range(iNumWeights):
			aiWeights.append(iDefaultWeight)
		
		# CP - move war checks here, counting number of civs we are at war with
		iWarCount = 0
		for iTeamLoop in range(gc.getMAX_CIV_TEAMS()):
			if (iTeamLoop != iTeam):
				pTeamLoop = gc.getTeam(iTeamLoop)
				if (pTeamLoop.isAlive() and not pTeamLoop.isBarbarian()):
					if (pTeamLoop.isAtWar(iTeam)):
						iWarCount += 1
		
		#######################################################################
		# HAPPINESS WEIGHTS
		#######################################################################
		
		printd("   Doing Happiness Weight")
		
		iAngryPop = pCity.angryPopulation(0)
		# AI Update for 1.65 - Only count angry people up to the system's current population limit,
		# angry people past that limit don't matter since they couldn't work anyway.
		# Note: if specialists are implemented then this adjustment should be removed
		# so that the actual angry population value is used (by removing the next two lines and this comment).
		iExcessPop = max(0, pCity.getPopulation() - pSystem.getPopulationLimit())
		iAngryPop = max(0, iAngryPop - iExcessPop)
		
		# Angry folks
		if (iAngryPop >= 5):
			aiWeights[iHappyWeightType] += 100 # CP - was 50
		if (iAngryPop >= 4):
			aiWeights[iHappyWeightType] += 100 # CP - was 50
		if (iAngryPop >= 3):
			aiWeights[iHappyWeightType] += 100 # CP - was 50
		if (iAngryPop >= 2):
			aiWeights[iHappyWeightType] += 100 # CP - was 50 
		if (iAngryPop >= 1):
			aiWeights[iHappyWeightType] += 150 # CP - was 115
		
		# Reduce for number of Sports Arenas already present
		# CP - count only those on planets within our current cultural borders
		# 	When a planet is captured, it can have them on planets outside our current reach
		iCount = 0	# AI Update for 1.65 - count the number that we could build, if there are none then the weight will be reduced to 0
		for iPlanetLoop in range(pSystem.getNumPlanets()):
			pPlanet = pSystem.getPlanetByIndex(iPlanetLoop)
			if (pPlanet.isPlanetWithinCulturalRange()) and pPlanet.isHasBuilding(iSportsArena) :
				aiWeights[iHappyWeightType] -= 15
			elif pPlanet.isPlanetWithinCulturalRange() :
				iCount += 1
			
		
		# CP - happy more important if not at war
		if (iWarCount == 0):
			aiWeights[iHappyWeightType] += 30
		
		# AI Update for 1.65 - if every planet currently within our reach already has a sports arena then there is
		# nothing we can do here, so drop the weight to 0.
		if (iCount == 0):
			aiWeights[iHappyWeightType] = 0
		
		#######################################################################
		# HEALTHINESS WEIGHTS
		#######################################################################
		
		printd("   Doing Healthiness Weight")
		
		iUnhealthy = pCity.badHealth(false) - pCity.goodHealth()
		
		# Unhealty folks
		if (iUnhealthy >= 5):
			aiWeights[iHealthWeightType] += 50
		if (iUnhealthy >= 4):
			aiWeights[iHealthWeightType] += 50
		if (iUnhealthy >= 3):
			aiWeights[iHealthWeightType] += 50
		if (iUnhealthy >= 2):
			aiWeights[iHealthWeightType] += 60
		if (iUnhealthy >= 1):
			aiWeights[iHealthWeightType] += 80
		
		# Reduce for number of Recycling Centers already present
		# CP - count only those on planets within our current cultural borders
		# 	When a planet is captured, it can have them on planets outside our current reach
		iCount = 0	# AI Update for 1.65 - count the number that we could build, if there are none then the weight will be reduced to 0
		for iPlanetLoop in range(pSystem.getNumPlanets()):
			pPlanet = pSystem.getPlanetByIndex(iPlanetLoop)
			if (pPlanet.isPlanetWithinCulturalRange()) and (pPlanet.isHasBuilding(iRecyclingCenter)):
				aiWeights[iHealthWeightType] -= 15
			elif pPlanet.isPlanetWithinCulturalRange() :
				iCount += 1
		
		# CP - healthy more important if not at war
		if (iWarCount == 0):
			aiWeights[iHealthWeightType] += 25
		
		# AI Update for 1.65 - if every planet currently within our reach already has a recycling center then there is
		# nothing we can do here, so drop the weight to 0.
		if (iCount == 0):
			aiWeights[iHealthWeightType] = 0
			
		
		#######################################################################
		# MILITARY WEIGHTS
		#######################################################################
		
		printd("   Doing Military Weight")
		
		# AI should really want to build Capital Shipyard if it doesn't already have one :)
		if (pPlayer.canConstruct(iCapitalShipyard, true, false, false)):	# Not actually sure what the middle 2 arguments do here
			if (pCity.getNumBuilding(iCapitalShipyard) == 0): # CP - this seems to have been left out, oops
				aiWeights[iMilitaryWeightType] += 80 #CP - bumped up from 60 to 80
		
		# CP - AI should want to build Squadron Factory if it doesn't already have one
		iSquadronFactory = pCivilization.getCivilizationBuildings(gc.getInfoTypeForString('BUILDINGCLASS_SQUADRON_FACTORY')) # this civ's SQUADRON_FACTORY
		if (pPlayer.canConstruct(iSquadronFactory, true, false, false)):	# Not actually sure what the middle 2 arguments do here
			if (pCity.getNumBuilding(iSquadronFactory) == 0): 
				aiWeights[iMilitaryWeightType] += 40
				
		# Presence of a Training Compound
		if (pCity.getNumBuilding(iTrainingCompound) > 0):
			aiWeights[iMilitaryWeightType] += 20
			
		iSchoolOfZealots = gc.getInfoTypeForString('BUILDING_SCHOOL_OF_ZEALOTS')
		# If a training compount makes it more militaristic, so should a school of zealots
		if (pCity.getNumBuilding(iSchoolOfZealots) > 0):
			aiWeights[iMilitaryWeightType] += 10
		
		# Current production level changes Military weight: 5 Prod gives +20, 10 Prod gives +40
		aiWeights[iMilitaryWeightType] += (pCity.getYieldRate(YieldTypes.YIELD_PRODUCTION) * 4)
		
		# War with other players: 50 weight each
		aiWeights[iMilitaryWeightType] += 50 * iWarCount

		if (iWarCount > 0):				
			# Also a chance of just quitting, letting the game do what it will during wartime
			iRand = CyGame().getSorenRandNum(100, "Final Frontier: Random roll to see if City AI override exits")
						
			if (iRand < (100 * (1.0 - (.6 ** iWarCount)))): # 40% per team we are at war with, like it was
				return false
		
		# Number of military units per city
		# CP - count missiles as half a unit for this
		iNumMilUnits = pPlayer.getNumMilitaryUnits()
		iNumMissiles = pPlayer.getUnitClassCount( gc.getInfoTypeForString('UNITCLASS_MISSILE_I'))
		iNumMissiles += pPlayer.getUnitClassCount( gc.getInfoTypeForString('UNITCLASS_MISSILE_II'))
		iNumMissiles += pPlayer.getUnitClassCount( gc.getInfoTypeForString('UNITCLASS_MISSILE_III'))
		iNumMissiles += pPlayer.getUnitClassCount( gc.getInfoTypeForString('UNITCLASS_MISSILE_Q'))
		iNumMilUnits -= ( iNumMissiles / 2 )
		# should not allow it to try to divide by zero...
		if (iNumMilUnits < 1) :
			iNumMilUnits = 1
		aiWeights[iMilitaryWeightType] += 100 * pPlayer.getNumCities() / iNumMilUnits
		
		
		#######################################################################
		# POPULATION WEIGHTS
		#######################################################################
		
		iPopToCap = pSystem.getPopulationLimit() - pCity.getPopulation()
		
		printd("   Doing Population Weight; iPopToCap is %d; PopLimit is %d" %(iPopToCap, pSystem.getPopulationLimit()))
		
		# Amount of Pop until this system hits its cap
		if (iPopToCap <= 1):
			aiWeights[iPopulationWeightType] += 60
		if (iPopToCap <= 0):
			aiWeights[iPopulationWeightType] += 60
		if (iPopToCap <= -1):
			aiWeights[iPopulationWeightType] += 60
		
		# Increase for City Size
		aiWeights[iPopulationWeightType] += (pCity.getPopulation() * 15) # CP - this weight factor seems a tad high, changing from 20 to 15
		
		# Reduce for number of Habitation Systems already present
		iNumHabitationSystems = pCity.getNumRealBuilding(iHabitationSystem)
		if (iNumHabitationSystems > 0):
			# CP - count only those on planets within our current cultural borders
			# 	When a planet is captured, it can have them on planets outside our current reach
			for iPlanetLoop in range(pSystem.getNumPlanets()):
				pPlanet = pSystem.getPlanetByIndex(iPlanetLoop)
				if (pPlanet.isPlanetWithinCulturalRange()) and pPlanet.isHasBuilding(iHabitationSystem) :
					aiWeights[iPopulationWeightType] -= 15
		
		
		
		#######################################################################
		# FOOD WEIGHTS
		#######################################################################
		
		printd("   Doing Food Weight")
		
		# Amount of food surplus in this city
		# CP - if the city is building with food in addition to production, then pCity.foodDifference returns 0
		# this seems like it may not be reset until you start building something that doesn't use food since
		# I am seeing some food weights that must have all of this tripped (totaling 220) even when the
		# city yield update in the log immediately prior to this is clearly 4 or more higher than the 
		# total population. So skip this when that is flagged too.
		if (not pPlayer.isAnarchy()) and (not  pCity.isFoodProduction()) :
			if (pCity.foodDifference(true) <= 3):
				aiWeights[iFoodWeightType] += 20 # CP - this was 40
			if (pCity.foodDifference(true) <= 2):
				aiWeights[iFoodWeightType] += 40 # CP - this was 50
			if (pCity.foodDifference(true) <= 1):
				aiWeights[iFoodWeightType] += 60
			if (pCity.foodDifference(true) <= 0):
				aiWeights[iFoodWeightType] += 70
		
		# Increase for City Size
		aiWeights[iFoodWeightType] += (pCity.getPopulation() * 15)
		
		# Decrease for number of Nutrition Facilities already present
		# CP - count only those on planets within our current cultural borders
		# 	When a system is captured, it can have them on planets outside our current reach
		iCount = 0	# AI Update for 1.65 - count the number that we could build, if there are none then the weight will be reduced to 0
		for iPlanetLoop in range(pSystem.getNumPlanets()):
			pPlanet = pSystem.getPlanetByIndex(iPlanetLoop)
			if (pPlanet.isPlanetWithinCulturalRange()) and pPlanet.isHasBuilding(iNutritionFacility) :
				aiWeights[iFoodWeightType] -= 10
			elif pPlanet.isPlanetWithinCulturalRange() :
				iCount += 1
				
		# AI Update for 1.65 - if every planet currently in reach already has a nutrition facility,
		# or if the current population is above the systems current population limit, then zero this weight.
		# Note: if specialists are implemented then the second condition should be removed so
		# the following line should be just "if (iCount == 0):"
		if ((iCount == 0) or ((pCity.getPopulation() - pSystem.getPopulationLimit()) > 0)):
			aiWeights[iFoodWeightType] = 0

			
		#######################################################################
		# PRODUCTION WEIGHTS
		#######################################################################
		
		printd("   Doing Production Weight")
		
		iProductionID = 1
		
		iDivisor = pCity.getBaseYieldRate(iProductionID)
		if (iDivisor == 0):
			iDivisor = 0.5
		
		aiWeights[iProductionWeightType] = int(150 * pCity.getPopulation() / iDivisor)		# Ex: 5 Base Production in 5 Pop System has weight of 150
		
		# Increase for City Size
		aiWeights[iProductionWeightType] += (pCity.getPopulation() * 10)
		
		# Decrease for number of Mining Facilities already present
		iNumMiningFacilities = pCity.getNumRealBuilding(iMiningFacility)
		if (iNumMiningFacilities > 0):
			# CP - count only those on planets within our current cultural borders
			# 	When a planet is captured, it can have them on planets outside our current reach
			for iPlanetLoop in range(pSystem.getNumPlanets()):
				pPlanet = pSystem.getPlanetByIndex(iPlanetLoop)
				if pPlanet.isPlanetWithinCulturalRange() and pPlanet.isHasBuilding(iMiningFacility) :
					aiWeights[iProductionWeightType] -= 10
		
		
		
		#######################################################################
		# COMMERCE WEIGHTS
		#######################################################################
		
		printd("   Doing Commerce Weight")
		
		iCommerceID = 2
		
		iDivisor = pCity.getBaseYieldRate(iCommerceID)
		if (iDivisor == 0):
			iDivisor = 0.5
		
		aiWeights[iCommerceWeightType] = int(250 * pCity.getPopulation() / iDivisor)		# Ex: 5 Base Commerce in 5 Pop System has weight of 250
		
		# Increase for City Size
		aiWeights[iCommerceWeightType] += (pCity.getPopulation() * 10)
		
		# Decrease for number of MagLev Networks already present
		iNumMagLevNetworks = pCity.getNumRealBuilding(iMagLevNetwork)
		if (iNumMagLevNetworks > 0):
			# CP - count only those on planets within our current cultural borders
			# 	When a planet is captured, it can have them on planets outside our current reach
			for iPlanetLoop in range(pSystem.getNumPlanets()):
				pPlanet = pSystem.getPlanetByIndex(iPlanetLoop)
				if pPlanet.isPlanetWithinCulturalRange() and pPlanet.isHasBuilding(iMagLevNetwork) :
					aiWeights[iCommerceWeightType] -= 20
		
		
		
		#######################################################################
		# Rank the weights
		#######################################################################
		
		# CP - reduce the military weight a little if we are not at war and there are other
		#      weights over 100.
		if (iWarCount == 0):
			for iWeightLoop in range(iNumWeights):
				if (iWeightLoop != iMilitaryWeightType) and (aiWeights[iWeightLoop] > 100) :
					aiWeights[iMilitaryWeightType] -= 15
		
		printd("   Ranking the Weights...")
		
		aaiWeightRankList = []
		
		# Create list
		for iWeightLoop in range(iNumWeights):
			aaiWeightRankList.append([aiWeights[iWeightLoop], iWeightLoop])
		
		# Order list by largest weight first
		aaiWeightRankList.sort()
		aaiWeightRankList.reverse()
		
		iBestWeight = aaiWeightRankList[0][0]
		iBestWeightType = aaiWeightRankList[0][1]
		
		for iWeightLoop in range(len(aaiWeightRankList)):
			printd("   WeightType: %s   Value: %d" %(aszWeightList[aaiWeightRankList[iWeightLoop][1]], aaiWeightRankList[iWeightLoop][0]))
		
		
		#######################################################################
		#######################################################################
		
		# Weights finished, now figure out what should be built
		
		#######################################################################
		#######################################################################

		# CP - If there is a bonus present on the system's plot and the bonus
		# related building has not been built yet, then maybe build it.
		pPlot = pCity.plot()
		iBonus = pPlot.getBonusType(iTeam)
		if ((iBonus != -1) and (pCity.getNumBuilding(FinalFrontier.dBonusBuildings[iBonus]) == 0)):
			# We have a bonus, but no bonus building. Determine if we will built it.
			# Odds for heath related bonus = 30% + 10% if aiWeights[iPopulationWeightType] > 150
			#	(iPopulationWeightType because these increase planet's pop limit)
			# Odds for happy related bonus = 25% + 10% if aiWeights[iCommerceWeightType] > 200
			#	(iCommerceWeightType because these increases commerce via per pop increase & trade routes)
			# The odds are a little lower for the happy bonuses because the regualr AI builds them more already.
			iRand = CyGame().getSorenRandNum(100, "FFP: Random roll to see if City AI will override for resource related building")
			iChance = 0
			if (gc.getBonusInfo(iBonus).getHealth() > 0):
				iChance = 30
				if (aiWeights[iPopulationWeightType] > 150):
					iChance += 10
			elif (gc.getBonusInfo(iBonus).getHappiness() > 0):
				iChance = 25
				if (aiWeights[iCommerceWeightType] > 200):
					iChance += 10
			
			if (iRand < iChance):
				# Force the build. To do this we need to first locate the planet with the bonus.
				for iPlanetLoop in range(pSystem.getNumPlanets()):
					pPlanet = pSystem.getPlanetByIndex(iPlanetLoop)
					if (pPlanet.isBonus()):
						# This is the place. Set this planet to be the current build planet
						# and push the order to build the appropriate building.
						printd("      Telling planet to build a planetary resource building (%d) at Ring %d" %(FinalFrontier.dBonusBuildings[iBonus],pPlanet.getOrbitRing()))
						pSystem.setBuildingPlanetRing(pPlanet.getOrbitRing())
						pCity.pushOrder(OrderTypes.ORDER_CONSTRUCT, FinalFrontier.dBonusBuildings[iBonus], -1, False, False, False, True)
						bOverride = true
						iBestWeight = 0 # make it skip the next section
						break
		
		# Must meet minimum weight to override
		if (iBestWeight >= iWeightNeeded):
			
			#######################################################################
			# Happiness
			#######################################################################
			
			if (iBestWeightType == iHappyWeightType):
				
				# Loop through the planets in the order of best to worst
				for iPlanetLoop in range(pSystem.getNumPlanets()):
					
					pPlanet = pSystem.getPlanetByIndex(iPlanetLoop)
					
					# Can only use planets in our radius
					if (pPlanet.isPlanetWithinCulturalRange()):
						
						# If we have a good enough food planet without a Habitation System (and can build one), then have the place construct one
						if (pPlayer.canConstruct(iSportsArena, true, false, false)):
							
							if (not pPlanet.isHasBuilding(iSportsArena)):
								
								printd("      Telling planet to build a Sports Arena at Ring %d" %(pPlanet.getOrbitRing()))
								
								pSystem.setBuildingPlanetRing(pPlanet.getOrbitRing())
								pCity.pushOrder(OrderTypes.ORDER_CONSTRUCT, iSportsArena, -1, False, False, False, True)
								bOverride = true
								break
			
			#######################################################################
			# Healthiness
			#######################################################################
			
			if (iBestWeightType == iHealthWeightType):
				
				# Loop through the planets in the order of best to worst
				for iPlanetLoop in range(pSystem.getNumPlanets()):
					
					pPlanet = pSystem.getPlanetByIndex(iPlanetLoop)
					
					# Can only use planets in our radius
					if (pPlanet.isPlanetWithinCulturalRange()):
						
						# If we have a good enough food planet without a Habitation System (and can build one), then have the place construct one
						if (pPlayer.canConstruct(iRecyclingCenter, true, false, false)):
							
							if (not pPlanet.isHasBuilding(iRecyclingCenter)):
								
								printd("      Telling planet to build a iRecyclingCenter at Ring %d" %(pPlanet.getOrbitRing()))
								
								pSystem.setBuildingPlanetRing(pPlanet.getOrbitRing())
								pCity.pushOrder(OrderTypes.ORDER_CONSTRUCT, iRecyclingCenter, -1, False, False, False, True)
								bOverride = true
								break
			
			#######################################################################
			# Military
			#######################################################################
			
			if (iBestWeightType == iMilitaryWeightType):
				
				if (pPlayer.canConstruct(iCapitalShipyard, true, false, false)):
					
					# No Capital Shipyard yet? BUILD IT!
					if (pCity.getNumBuilding(iCapitalShipyard) == 0):
						# CP - well, perhaps not. This is one of my AI improvements.
						# If the city has a total production yield of 10+ then force it,
						# if it is 7 to 9 then a 80% chance to force it,
						# if it is 3 to 6 then a 20% chance to force it,
						# if it is 2 or less then don't force it.
						iProd = pCity.getYieldRate( YieldTypes.YIELD_PRODUCTION )
						bForceBuild = false
						if ( iProd >= 10 ) :
							bForceBuild = true
						elif ( iProd >= 7 ) :
							iRand = CyGame().getSorenRandNum(100, "Finaler Frontier: Random roll to see if City AI will override for Capital Shipyard (1)")
							if ( iRand < 80):
								bForceBuild = true
						elif ( iProd >= 3 ) :
							iRand = CyGame().getSorenRandNum(100, "Finaler Frontier: Random roll to see if City AI will override for Capital Shipyard (2)")
							if ( iRand < 20):
								bForceBuild = true
						
						if ( bForceBuild ) :
							printd("      Telling System to build a Capital Shipyard")
							pCity.pushOrder(OrderTypes.ORDER_CONSTRUCT, iCapitalShipyard, -1, False, False, False, True)
							bOverride = true
			
				# CP - AI improvement (hopefully) (the indentation level in this gets a tad excessive)
				# If we are not overriding then either we can't build a Capitol Shipyard or we already have one.
				# If we don't have one then don't do anything here.
				# If we do have one, then check for Squadron Factory.
				# If we can build one and don't have one then 75% to build one (otherwise return without overriding)
				# If we have a squadron factory:
				# Find out if we are using the Squadron Doctrine civic.
				# Now we are set to do some other checking to see if we should force some builds
				#	If we can train the Carrier unitclass
				#		If we have no carriers then 75% to force one, 90% if in Squadron Doctrine
				#		If we have 1 carrier then 10% to force another, 50% if in Squadron Doctrine
				#		If we have >1 carriers then 0% to force another, 10% if in Squadron Doctrine
				# If we are not forcing a carrier to be built, check if we have enough fighters
				#	If we are not in Squadron Doctrine, enough is 1/2 system count + number of carriers
				#	If we are in Squadron Doctrine, enough is as above + 2
				#	If we don't have enough, then 50% chance to force training one, 75% if in Squadron Doctrine
				# If we are not forcing anything to be ebuilt yet, check if we have enough bombers
				#	If we are not in Squadron Doctrine, enough is 1/2 system count + (2 * number of carriers)
				#	If we are in Squadron Doctrine, enough is as above + 2
				#	If we don't have enough, then 50% chance to force training one, 75% if in Squadron Doctrine
				#
				# Note that all through here we are making sure we use UnitClass -> Unit for this civ to get
				# the right unit built (although, at this point, nobody has a UU for any of these except
				# the bomber squadrons).
				if ((not bOverride) and ( pCity.getNumBuilding(iCapitalShipyard) > 0 ) ) :
					# We will only attempt to force a squadron factory if we already have a capital shipyard
					# (forced or not). The AI will build a few squadron factories on it's own, as it does end
					# up building some fighters and bombers without this change, so I don't feel the need
					# to force them so much. A side effect of the way the code is constructed is that it
					# will also only try to force a carrier, fighter, or bomber if there is already a
					# a capital shipyard and a squadron factory.
					iNumCarriers = 0
					if (pPlayer.canConstruct( iSquadronFactory, true, false, false)):
						if (pCity.getNumBuilding(iSquadronFactory) == 0):
							
							iRand = CyGame().getSorenRandNum(100, "Finaler Frontier: Random roll to see if City AI will override for Squadron Factory")
							if (iRand < 75):
								# 75% chance to force one to be built
								
								printd("      Telling System to build a Squadron Factory")
								pCity.pushOrder(OrderTypes.ORDER_CONSTRUCT, iSquadronFactory, -1, False, False, False, True)
								bOverride = true
				
						else :
							# The city has a squadron factory. Before we check for forcing fighters or bombers
							# I'm checking to see if we want to force a carrier.
							
							iCarrierI = gc.getInfoTypeForString('UNITCLASS_CARRIER_I')
							iCarrierII = gc.getInfoTypeForString('UNITCLASS_CARRIER_II')
							iCarrierIII = gc.getInfoTypeForString('UNITCLASS_CARRIER_III')
							
							iNumUnits = pPlayer.getUnitClassCountPlusMaking( iCarrierI)
							iNumUnits += pPlayer.getUnitClassCountPlusMaking( iCarrierII)
							iNumUnits += pPlayer.getUnitClassCountPlusMaking( iCarrierIII)
							iNumCarriers = iNumUnits							
							iSquadronDoctrine = gc.getInfoTypeForString('CIVIC_SQUADRON_DOCTRINE')
							bIsSquadronDoctrine = pPlayer.isCivic(iSquadronDoctrine)
														
							iCarrierIUnit = pCivilization.getCivilizationUnits( iCarrierI ) # get this civilization's unit for this unitclass
							
							if ( pPlayer.canTrain( iCarrierIUnit, true, false) ) :
								# It is important to note that we can only get here if the city already has
								# a Capital Shipyard (otherwise we'd have to check for one)
															
								bForceBuild = false
								if (iNumUnits == 0) :
									iRand = CyGame().getSorenRandNum(100, "Finaler Frontier: Random roll to see if City AI will override for carrier (1)")
									if ( bIsSquadronDoctrine ) :
										if ( iRand < 80):
											# 90% chance to force one ## until a fix to get them to actually carry squadrons reduce to 80
											bForceBuild = true
									else :
										if ( iRand < 0):
											# 75% chance to force one ## until a fix to get them to actually carry squadrons reduce to 0
											bForceBuild = true
							
								elif (iNumUnits == 1) :
									iRand = CyGame().getSorenRandNum(100, "Finaler Frontier: Random roll to see if City AI will override for carrier (2)")
									if ( bIsSquadronDoctrine ) :
										if ( iRand < 10):
											# 50% chance to force one ## until a fix to get them to actually carry squadrons reduce to 10
											bForceBuild = true
									else :
										if ( iRand < 0):
											# 10% chance to force one ## until a fix to get them to actually carry squadrons reduce to 0
											bForceBuild = true
							
								else :
									# iNumUnits must be > 1
									iRand = CyGame().getSorenRandNum(100, "Finaler Frontier: Random roll to see if City AI will override for carrier (3)")
									if ( bIsSquadronDoctrine ) :
										if ( iRand < 10):
											# 10% chance to force one
											bForceBuild = true
								
								if ( bForceBuild ) :
									iCarrierIIUnit = pCivilization.getCivilizationUnits( iCarrierII ) # get this civilization's unit for this unitclass
									iCarrierIIIUnit = pCivilization.getCivilizationUnits( iCarrierIII ) # get this civilization's unit for this unitclass
							
									# Build the most advanced one we can build, of course
									iBuildUnit = iCarrierIUnit
									if ( pPlayer.canTrain( iCarrierIIUnit, true, false) ) :
										iBuildUnit = iCarrierIIUnit
									if ( pPlayer.canTrain( iCarrierIIIUnit, true, false) ) :
										iBuildUnit = iCarrierIIIUnit
							
									printd("      Telling System to build a Carrier")
									pCity.pushOrder(OrderTypes.ORDER_TRAIN, iBuildUnit, -1, False, False, False, True)
									bOverride = true
							
							# Check for forcing the training of a fighter unit
							iSpaceFighterI = gc.getInfoTypeForString('UNITCLASS_SPACE_FIGHTER_I')
							iSpaceFighterIUnit = pCivilization.getCivilizationUnits( iSpaceFighterI ) # get this civilization's unit for this unitclass
							iSquadronQ = gc.getInfoTypeForString('UNITCLASS_SQUADRON_Q')
							iSquadronQUnit = pCivilization.getCivilizationUnits( iSquadronQ ) # get this civilization's unit for this unitclass
																		
							if ((not bOverride) and ( pPlayer.canTrain( iSpaceFighterIUnit, true, false) ) ) :
								# Check to see if we have enough fighters
								iSpaceFighterII = gc.getInfoTypeForString('UNITCLASS_SPACE_FIGHTER_II')
								iSpaceFighterIII = gc.getInfoTypeForString('UNITCLASS_SPACE_FIGHTER_III')

								iEnoughUnits = (pPlayer.getNumCities() / 2) + iNumCarriers
								if ( bIsSquadronDoctrine ) :
									iEnoughUnits += 2
				
								iNumUnits = pPlayer.getUnitClassCountPlusMaking( iSpaceFighterI)
								iNumUnits += pPlayer.getUnitClassCountPlusMaking( iSpaceFighterII)
								iNumUnits += pPlayer.getUnitClassCountPlusMaking( iSpaceFighterIII)
								iNumUnits += pPlayer.getUnitClassCountPlusMaking( iSquadronQ)
								
								bForceBuild = false
								if ( iNumUnits < iEnoughUnits ) :
									iRand = CyGame().getSorenRandNum(100, "Finaler Frontier: Random roll to see if City AI will override for fighter")
									if ( bIsSquadronDoctrine ) :
										if ( iRand < 75):
											# 75% chance to force one
											bForceBuild = true
									else :
										if ( iRand < 50):
											# 50% chance to force one
											bForceBuild = true
							
									if ( bForceBuild ) :
										iSpaceFighterIIUnit = pCivilization.getCivilizationUnits( iSpaceFighterII ) # get this civilization's unit for this unitclass
										iSpaceFighterIIIUnit = pCivilization.getCivilizationUnits( iSpaceFighterIII ) # get this civilization's unit for this unitclass
							
										# Build the most advanced one we can build, of course
										iBuildUnit = iSpaceFighterIUnit
										if ( pPlayer.canTrain( iSpaceFighterIIUnit, true, false) ) :
											iBuildUnit = iSpaceFighterIIUnit
										if ( pPlayer.canTrain( iSpaceFighterIIIUnit, true, false) ) :
											iBuildUnit = iSpaceFighterIIIUnit
										if ( pPlayer.canTrain( iSquadronQUnit, true, false) ) :
											iBuildUnit = iSquadronQUnit
							
										printd("      Telling System to build a fighter")
										pCity.pushOrder(OrderTypes.ORDER_TRAIN, iBuildUnit, -1, False, False, False, True)
										bOverride = true
							
							# Check for forcing the training of a bomber unit
							iSpaceBomberI = gc.getInfoTypeForString('UNITCLASS_SPACE_BOMBER_I')
							iSpaceBomberIUnit = pCivilization.getCivilizationUnits( iSpaceBomberI ) # get this civilization's unit for this unitclass
							
							if ((not bOverride) and ( pPlayer.canTrain( iSpaceBomberIUnit, true, false) ) ) :
								# Check to see if we have enough fighters
								iSpaceBomberII = gc.getInfoTypeForString('UNITCLASS_SPACE_BOMBER_II')
								iSpaceBomberIII = gc.getInfoTypeForString('UNITCLASS_SPACE_BOMBER_III')
								
								iEnoughUnits = (pPlayer.getNumCities() / 2) + ( 2 * iNumCarriers)
								if ( bIsSquadronDoctrine ) :
									iEnoughUnits += 2
				
								iNumUnits = pPlayer.getUnitClassCountPlusMaking( iSpaceBomberI)
								iNumUnits += pPlayer.getUnitClassCountPlusMaking( iSpaceBomberII)
								iNumUnits += pPlayer.getUnitClassCountPlusMaking( iSpaceBomberIII)
								iNumUnits += pPlayer.getUnitClassCountPlusMaking( iSquadronQ)
							
								bForceBuild = false
								if ( iNumUnits < iEnoughUnits ) :
									iRand = CyGame().getSorenRandNum(100, "Finaler Frontier: Random roll to see if City AI will override for bomber")
									if ( bIsSquadronDoctrine ) :
										if ( iRand < 75):
											# 75% chance to force one
											bForceBuild = true
									else :
										if ( iRand < 50):
											# 50% chance to force one
											bForceBuild = true
							
									if ( bForceBuild ) :
										iSpaceBomberIIUnit = pCivilization.getCivilizationUnits( iSpaceBomberII ) # get this civilization's unit for this unitclass
										iSpaceBomberIIIUnit = pCivilization.getCivilizationUnits( iSpaceBomberIII ) # get this civilization's unit for this unitclass
							
										# Build the most advanced one we can build, of course
										iBuildUnit = iSpaceBomberIUnit
										if ( pPlayer.canTrain( iSpaceBomberIIUnit, true, false) ) :
											iBuildUnit = iSpaceBomberIIUnit
										if ( pPlayer.canTrain( iSpaceBomberIIIUnit, true, false) ) :
											iBuildUnit = iSpaceBomberIIIUnit
										if ( pPlayer.canTrain( iSquadronQUnit, true, false) ) :
											iBuildUnit = iSquadronQUnit
							
										printd("      Telling System to build a bomber")
										pCity.pushOrder(OrderTypes.ORDER_TRAIN, iBuildUnit, -1, False, False, False, True)
										bOverride = true
							
			#######################################################################
			# Population
			#######################################################################
			
			if (iBestWeightType == iPopulationWeightType):
				
				iFood = 0
				
				# Find the best planet to give us some food
				aiSizeFoodPlanetIndexList = pSystem.getSizeYieldPlanetIndexList(iFood) # CP- take planet size into consideration
				
				iPass = 0
				
				# Loop through the planets in the order of best to worst
				for iPlanetLoop in range(len(aiSizeFoodPlanetIndexList)):
					
					iPlanetFromList = aiSizeFoodPlanetIndexList[iPlanetLoop][2]	# CP - the list is lists of pop limit + food value + planet index
					pPlanet = pSystem.getPlanetByIndex(iPlanetFromList)		#    - select in descending order of "size" food-per-pop
					
					# Can only use planets in our radius
					if (pPlanet.isPlanetWithinCulturalRange()):
						
						# First try for an extended habitation system
						if (pPlayer.canConstruct(iExtHabitationSystem, true, false, false)):
							
							# Doesn't already have building
							if ((not pPlanet.isHasBuilding(iExtHabitationSystem)) and pPlanet.isHasBuilding(iHabitationSystem) and (iPass < pCity.getPopulation())):
								
								printd("      Telling planet to build an Exttended Habitation System at Ring %d" %(pPlanet.getOrbitRing()))
								
								pSystem.setBuildingPlanetRing(pPlanet.getOrbitRing())
								pCity.pushOrder(OrderTypes.ORDER_CONSTRUCT, iExtHabitationSystem, -1, False, False, False, True)
								bOverride = true
								break
								
							# Increment if player CANNOT build a Habitation System (this is not actually possible given the existing tech tree)
							elif (not pPlayer.canConstruct(iHabitationSystem, true, false, false)):
								iPass += 1
								
						# If we have a good enough food planet without a Habitation System (and can build one), then have the place construct one
						if (pPlayer.canConstruct(iHabitationSystem, true, false, false)):
							
							# Good food on this planet, good for future growth
# CP - the following line makes no sense, so comment it out...
#							if (pPlanet.getTotalYield(iPlayer, iFood) >= 3):
								
								# Doesn't already have building
								if (not pPlanet.isHasBuilding(iHabitationSystem) and iPass < pCity.getPopulation()):
									
									printd("      Telling planet to build a Habitation System at Ring %d" %(pPlanet.getOrbitRing()))
									
									pSystem.setBuildingPlanetRing(pPlanet.getOrbitRing())
									pCity.pushOrder(OrderTypes.ORDER_CONSTRUCT, iHabitationSystem, -1, False, False, False, True)
									bOverride = true
									break
									
								else:
									iPass += 1
			
			#######################################################################
			# Food
			#######################################################################
			
			if (iBestWeightType == iFoodWeightType):
				
				iFood = 0
				
				# Find the best planet to give us some food
				aiSizeFoodPlanetIndexList = pSystem.getSizeYieldPlanetIndexList(iFood)
				
				iPass = 0
				
				# Loop through the planets in the order of best to worst (for food)
				for iPlanetLoop in range(len(aiSizeFoodPlanetIndexList)):
					
					iPlanetFromList = aiSizeFoodPlanetIndexList[iPlanetLoop][2]	# CP - the list is lists of pop limit + food value + planet index
					pPlanet = pSystem.getPlanetByIndex(iPlanetFromList)			#    - select in descending order of "size" and food-per-pop
					
					# Can only use planets in our radius
					if (pPlanet.isPlanetWithinCulturalRange()):
						
						if (pPlayer.canConstruct(iNutritionFacility, true, false, false)):
							
							# Doesn't already have building
							if (not pPlanet.isHasBuilding(iNutritionFacility) and iPass < pCity.getPopulation()):
								
								printd("      Telling planet to build a iNutritionFacility at Ring %d" %(pPlanet.getOrbitRing()))
								
								pSystem.setBuildingPlanetRing(pPlanet.getOrbitRing())
								pCity.pushOrder(OrderTypes.ORDER_CONSTRUCT, iNutritionFacility, -1, False, False, False, True)
								bOverride = true
								break
								
							else:
								iPass += 1
			
			#######################################################################
			# Production
			#######################################################################
			
			if (iBestWeightType == iProductionWeightType):
				
				iFood = 0	 #Add to our food planets :)
				
				# Find the best planet to give us some Production
				aiSizeFoodPlanetIndexList = pSystem.getSizeYieldPlanetIndexList(iFood)
				
				iPass = 0
				
				# Loop through the planets in the order of best to worst (for Food)
				for iPlanetLoop in range(len(aiSizeFoodPlanetIndexList)):
					
					iPlanetFromList = aiSizeFoodPlanetIndexList[iPlanetLoop][2]	# CP - the list is lists of pop limit + food value + planet index
					pPlanet = pSystem.getPlanetByIndex(iPlanetFromList)			#    - select in descending order of "size" and food-per-pop
					
					# Can only use planets in our radius
					if (pPlanet.isPlanetWithinCulturalRange()):
						
						# First try for a nanoextraction upgrade system
						if (pPlayer.canConstruct(iNanoExtraction, true, false, false)):
							
							# Doesn't already have building
							if ((not pPlanet.isHasBuilding(iNanoExtraction)) and pPlanet.isHasBuilding(iMiningFacility) and (iPass < pCity.getPopulation())):
								
								printd("      Telling planet to build a nanoextraction upgrade at Ring %d" %(pPlanet.getOrbitRing()))
								
								pSystem.setBuildingPlanetRing(pPlanet.getOrbitRing())
								pCity.pushOrder(OrderTypes.ORDER_CONSTRUCT, iNanoExtraction, -1, False, False, False, True)
								bOverride = true
								break
								
							# Increment if player CANNOT build a Mining Facility (this is not actually possible given the existing tech tree)
							elif (not pPlayer.canConstruct(iMiningFacility, true, false, false)):
								iPass += 1

						if (pPlayer.canConstruct(iMiningFacility, true, false, false)):
							
							# Doesn't already have building
							if (not pPlanet.isHasBuilding(iMiningFacility) and iPass < pCity.getPopulation()):
								
								printd("      Telling planet to build a iMiningFacility at Ring %d" %(pPlanet.getOrbitRing()))
								
								pSystem.setBuildingPlanetRing(pPlanet.getOrbitRing())
								pCity.pushOrder(OrderTypes.ORDER_CONSTRUCT, iMiningFacility, -1, False, False, False, True)
								bOverride = true
								break
								
							else:
								iPass += 1
						
			#######################################################################
			# Commerce
			#######################################################################
			
			if (iBestWeightType == iCommerceWeightType):
				
				iFood = 0	 #Add to our food planets :)
				
				# Find the best planet to give us some Commerce
				aiSizeFoodPlanetIndexList = pSystem.getSizeYieldPlanetIndexList(iFood)
				
				iPass = 0
				
				# Loop through the planets in the order of best to worst (for Food)
				for iPlanetLoop in range(len(aiSizeFoodPlanetIndexList)):
					
					iPlanetFromList = aiSizeFoodPlanetIndexList[iPlanetLoop][2]	# CP - the list is lists of pop limit + food value + planet index
					pPlanet = pSystem.getPlanetByIndex(iPlanetFromList)			#    - select in descending order of "size" and food-per-pop
					
					# Can only use planets in our radius
					if (pPlanet.isPlanetWithinCulturalRange()):
						
						# Commercial Satellites
						if (pPlayer.canConstruct(iCommercialSatellites, true, false, false)):
							
							# Doesn't already have building
							if ((not pPlanet.isHasBuilding(iCommercialSatellites)) and pPlanet.isHasBuilding(iMagLevNetwork) and (iPass < pCity.getPopulation())): # CP - bug fix: check for mag-lev before trying to construct com. sat.
								
								printd("      Telling planet to build a iCommercialSatellites at Ring %d" %(pPlanet.getOrbitRing()))
								
								pSystem.setBuildingPlanetRing(pPlanet.getOrbitRing())
								pCity.pushOrder(OrderTypes.ORDER_CONSTRUCT, iCommercialSatellites, -1, False, False, False, True)
								bOverride = true
								break
								
							# Increment if player CANNOT build a MagLev
							elif (not pPlayer.canConstruct(iMagLevNetwork, true, false, false)):
								iPass += 1
								
						# Can't build Commercial Satellites, go for MagLev instead
						if (pPlayer.canConstruct(iMagLevNetwork, true, false, false)):
							
							# Doesn't already have building
							if (not pPlanet.isHasBuilding(iMagLevNetwork) and iPass < pCity.getPopulation()):
								
								printd("      Telling planet to build a iMagLevNetwork at Ring %d" %(pPlanet.getOrbitRing()))
								
								pSystem.setBuildingPlanetRing(pPlanet.getOrbitRing())
								pCity.pushOrder(OrderTypes.ORDER_CONSTRUCT, iMagLevNetwork, -1, False, False, False, True)
								bOverride = true
								break
								
							else:
								iPass += 1
					
				
				
		
		
		return bOverride
		
		
		
		
		
		
		
		
##########################################################
##########################################################
##########################################################

#		UNIT UPDATE

##########################################################
##########################################################
##########################################################
		
	def doStarbaseAI(self, pUnit):
		
		if (not gc.getPlayer(pUnit.getOwner()).isHuman()):
			
			printd("\n\n\n\n\nDoing AI for %s at (%d, %d)" %(pUnit.getName(), pUnit.getX(), pUnit.getY()))
			
			iOwner = pUnit.getOwner()
			pOwner = gc.getPlayer(iOwner)
			iTeam = pOwner.getTeam()
			pTeam = gc.getTeam(iTeam)
			
			aiBombardPlotValues = []
			
			# Check plots within range to see what's around us
			for iXLoop in range(pUnit.getX()-pUnit.airRange(), pUnit.getX()+pUnit.airRange()+1):
				for iYLoop in range(pUnit.getY()-pUnit.airRange(), pUnit.getY()+pUnit.airRange()+1):
					
					iActiveX = iXLoop
					iActiveY = iYLoop
					
					if (iActiveX >= CyMap().getGridWidth()):
						iActiveX = iActiveX - CyMap().getGridWidth()
					if (iActiveY >= CyMap().getGridHeight()):
						iActiveY = iActiveY - CyMap().getGridHeight()
					if (iActiveX < 0):
						iActiveX = CyMap().getGridWidth() + iActiveX
					if (iActiveY < 0):
						iActiveY = CyMap().getGridHeight() + iActiveY
					
					pLoopPlot = CyMap().plot(iActiveX, iActiveY)
					
					#AI fixes by God-Emperor...
					if pLoopPlot.isVisible(iTeam, false) and (plotDistance(pUnit.getX(),pUnit.getY(),iActiveX,iActiveY) <= pUnit.airRange()): # CP - bugfix for AI starbases attacking things they shouldn't be able to see or reach
						iPlotValue = 0
					
						# Look at all the units on this plot
						for iUnitLoop in range(pLoopPlot.getNumUnits()):
						
							pLoopUnit = pLoopPlot.getUnit(iUnitLoop)
						
							# At war with this unit's owner?
							if (pTeam.isAtWar(pLoopUnit.getTeam())):
							
								# The greater the cost of the unit the more we want to kill it :)
								iCost = gc.getUnitInfo(pLoopUnit.getUnitType()).getProductionCost()
							
								if (iCost > 0):
									iPlotValue += iCost
									
						if iPlotValue > 0: # CP - bugfix for trying to attack when there are no valid targets
							aiBombardPlotValues.append([iPlotValue, iActiveX, iActiveY])
					
			# Any valid plots to hit?
			if (len(aiBombardPlotValues) > 0):
				
				# Order list of plots based on most valuable first (first element in the list)
				aiBombardPlotValues.sort()
				aiBombardPlotValues.reverse()
				
				# Now attack! :)
				
				aiBombardPlotList = aiBombardPlotValues[0]
				
				iX = aiBombardPlotList[1]
				iY = aiBombardPlotList[2]
				
				pUnit.rangeStrike(iX, iY)
				
	def doConstructionShipAI(self, pUnit):
		
		printd("\n\n\n\n\nDoing AI for %s at (%d, %d)" %(pUnit.getName(), pUnit.getX(), pUnit.getY()))
		
		bOverride = false
		
		iBuild = gc.getInfoTypeForString('BUILD_STARBASE')
		
		pPlayer = gc.getPlayer(pUnit.getOwner())
		pTeam = gc.getTeam(pPlayer.getTeam())
		
		# Is there a valid Starbase maker assigned?
		pPlayerAIInfo = self.getPlayerAIInfo(pUnit.getOwner())
		
		iMakerID = pPlayerAIInfo.getUnitIDStarbaseMaker()
		pMakingUnit = pPlayer.getUnit(iMakerID)
		
		bUnitExists = false
		if (pMakingUnit):
			if (pMakingUnit.getName()):
				bUnitExists = true
			
		if (iMakerID == -1 or not bUnitExists):
			pPlayerAIInfo.setUnitIDStarbaseMaker(pUnit.getID())
		
		printd("  Unit we want: %d, Unit we have: %d" %(pPlayerAIInfo.getUnitIDStarbaseMaker(), pUnit.getID()))
		
		# This is the right unit to send...
		if (pPlayerAIInfo.getUnitIDStarbaseMaker() == pUnit.getID()):
			
			printd("    This is the unit we want, trying to Override unit AI")
			
			# Enough gold to do?
			iBuildCost = gc.getBuildInfo(iBuild).getCost() * (100 + pPlayer.calculateInflationRate()) / 100
			
			printd("    Build cost for Starbase is %d" %(iBuildCost))
			
			if (pPlayer.getGold() > iBuildCost):
				
				printd("      Player has enough gold (%d)" %(iBuildCost))
				
#				if (pPlayer.canBuild(CyPlot* pPlot, iBuild, false, false)):
				
				# Has the right techs?
				iTech = gc.getInfoTypeForString('TECH_INDUSTRY_1')
				if (pTeam.isHasTech(iTech)):
					
					printd("         Player has right techs")
					
					# This function is somewhat expensive... should probably call it less frequently
					self.doStarbaseNeed(pUnit.getOwner())
					
					# Should we be trying to send a ship to build a Starbase?
					if (pPlayerAIInfo.getStarbaseTargetValue() > 900):
						
						printd("          Plot value is high enough to send someone to build a Starbase")
						
						printd("            Starbase Target: %s" %(pPlayerAIInfo.getStarbaseTarget()))
						printd("            Unit XY: %d, %d" %(pUnit.getX(), pUnit.getY()))
						
						# Already in the right location?
						iX, iY = pPlayerAIInfo.getStarbaseTarget()
						if (pUnit.getX() == iX and pUnit.getY() == iY):
							
							printd("          Unit at the right location to build Starbase")
							
							# Not already building Starbase?
							if (pUnit.getBuildType() != iBuild):
								
								# Tell unit to build Starbase
								printd("            Telling unit to build Starbase")
								# Can build?
								if (pUnit.canBuild(pUnit.plot(), iBuild, true)):
									
									bCanDo = true
									
									# Loop through all units on the plot, if there are any Construction Ships already building a Starbase then disallow more	
									for iUnitLoop in range(pUnit.plot().getNumUnits()):
										pUnitLoop = pUnit.plot().getUnit(iUnitLoop)
										if (pUnitLoop.getBuildType() == gc.getInfoTypeForString('BUILD_STARBASE')):
											bCanDo = false
									
									if (bCanDo):
										printd("              Able to build Starbase, pushing mission & setting override")
										pUnit.getGroup().pushMission(MissionTypes.MISSION_BUILD, iBuild, -1, -1, false, true, MissionAITypes.MISSIONAI_BUILD, pUnit.plot(), pUnit)
										bOverride = true
						
						# Unit not in the right spot, send him in the right direction
						else:
							printd("          Unit isn't in the right spot yet, so we're going to tell him to move there")
							if (pUnit.canMoveInto(CyMap().plot(iX, iY), false, false, false)):
								printd("            And we did")
								pUnit.getGroup().pushMoveToMission(iX, iY)
								pUnit.finishMoves()
								bOverride = true
		
		# Need more gold!
		else:
			pPlayer.AI_setExtraGoldTarget(250)
							
						
		printd("Override unit AI? : %s\n\n\n" %(bOverride))
		return bOverride
		
	def doStarbaseNeed(self, iPlayer):
		
		# Loop through all plots, see if there are any really good places to get resources
		
		pBestPlot, iBestValue = self.findBestResourcePlot(iPlayer, true)
		
		printd("\npBestPlot")
		printd(pBestPlot)
		printd("iBestValue")
		printd(iBestValue)
		
		if (pBestPlot != -1):
			printd("   X: %d Y: %d" %(pBestPlot.getX(), pBestPlot.getY()) )
			
			# 700 is our cutoff point for a good value
			if (iBestValue >= 700):
				
				pPlayerAIInfo = self.getPlayerAIInfo(iPlayer)
				
				# Set the player's current Starbase destination
				pPlayerAIInfo.setStarbaseTarget(pBestPlot.getX(), pBestPlot.getY())
				pPlayerAIInfo.setStarbaseTargetValue(iBestValue)
		
	def findBestResourcePlot(self, iPlayer=-1, bCheckForVisibility=false):
		
		pBestPlot = -1
		iBestValue = -1
		
		pPlayer = -1
		iTeam = -1
		if (iPlayer >= 0):
			pPlayer = gc.getPlayer(iPlayer)
			iTeam = pPlayer.getTeam()
		
		printd("\n    Checking for best plot to build a Starbase \n")
		
		for iPlotLoop in range(CyMap().numPlots()):
			pLoopPlot = CyMap().plotByIndex(iPlotLoop)
#			printd("Checking value of plot at %d, %d" %(pLoopPlot.getX(), pLoopPlot.getY()))
			
			# If we're supposed to be checking for a player's visibility then only check this plot if it's revealed
			if (bCheckForVisibility):
				if (not pLoopPlot.isRevealed(iTeam, false)):
					continue

			# CP - Check the plot being rated to see if it already belongs to someone else.
			iPlotOwner = pLoopPlot.getOwner()
			if ((iPlotOwner != -1) and (iPlotOwner != iPlayer)):
				continue
			
			iDistanceFromCapital = CyMap().getGridWidth()
			
			if (pPlayer.getCapitalCity()):
				iDistanceFromCapital = CyMap().calculatePathDistance(pPlayer.getCapitalCity().plot(), pLoopPlot)
			
			# Don't look too far away (performance, more than anything)
			iMaxRange = max(CyMap().getGridWidth() / 2, 60)
			if (iDistanceFromCapital > 0 and iDistanceFromCapital < iMaxRange):
				
				# Will be a value between 0 and the width of the map * 100, e.g. 64 * 100 (6400)
				iDistanceValueMod = (CyMap().getGridWidth() * 100) / iDistanceFromCapital
				iDistanceValueMod = math.sqrt(iDistanceValueMod) * 10
				
				iPlotValue = 0
				iNumBonuses = 0
				
				for iXSearchLoop in range(pLoopPlot.getX()-2, pLoopPlot.getX()+3):
					for iYSearchLoop in range(pLoopPlot.getY()-2, pLoopPlot.getY()+3):
						pSearchPlot = CyMap().plot(iXSearchLoop, iYSearchLoop)
						
						# Don't search unseen plots in range of the one we're looking at either
						if (bCheckForVisibility):
							if (not pSearchPlot.isRevealed(iTeam, false)):
								continue
						
						# Bonus present?
						if (pSearchPlot.getBonusType(iTeam) != -1):
							# Bonus unowned?
							if (pSearchPlot.getOwner() == -1):
								iNumBonuses += 1

						#Build starbases close to solar systems (for defensive purposes) -- TC01
						if pSearchPlot.isCity():
							if pSearchPlot.getPlotCity().getOwner() == iPlayer:
								iPlotValue += 700		#Too much?

						#Build starbases near wormholes (since AI doesn't get Wormholes) -- TC01
						if pSearchPlot.getFeatureType() != -1:
							if gc.getFeatureInfo(pSearchPlot.getFeatureType()).getTargetWormholeType() != -1:
								iPlotValue += 700
							
						#If other starbases are present in radius, no starbase -- TC01
						for iUnit in range(pSearchPlot.getNumUnits()):
							pOtherStarbase = pSearchPlot.getUnit(iUnit)
							if pOtherStarbase.isStarbase():
								iPlotValue = 0
								iNumBonuses = 0
								break

				if (iNumBonuses > 0):
					iBonusValueMod = ((1280 / 3) * iNumBonuses)
					iPlotValue += iDistanceValueMod + iBonusValueMod
					printd("      Plot value for (%d, %d) is %d: Distance: %d, Bonus: %d" %(pLoopPlot.getX(), pLoopPlot.getY(), iPlotValue, iDistanceValueMod, iBonusValueMod))
				
				# Don't build anywhere except in empty space & asteroids
				if (pLoopPlot.getFeatureType() != -1 and pLoopPlot.getFeatureType() != gc.getInfoTypeForString('FEATURE_FOREST')):
					iPlotValue = 0
					
				# Little extra bonus for being in Asteroids (defense)
				if (pLoopPlot.getFeatureType() == gc.getInfoTypeForString('FEATURE_FOREST')):
					iPlotValue += 50		#Formerly 15... asteroids are good! -- TC01
					
				# If this plot has the most resources in range from what we've found
				if (iPlotValue > iBestValue):
					iBestValue = iPlotValue
					pBestPlot = pLoopPlot
			
		return [pBestPlot, iBestValue]
		
class CvPlayerAIInfo:
	
	def __init__(self, iID):
		
		self.iID = iID
		
		self.iStarbaseTargetValue = 0
		
		self.iXStarbaseTarget = -1
		self.iYStarbaseTarget = -1
		
		self.iUnitIDStarbaseMaker = -1
		
	def getID(self):
		return self.iID
	def setID(self, iValue):
		self.iID = iValue
		
	def getStarbaseTargetValue(self):
		return self.iStarbaseTargetValue
	def setStarbaseTargetValue(self, iValue):
		self.iStarbaseTargetValue = iValue

	def getStarbaseTarget(self):
		return [self.iXStarbaseTarget, self.iYStarbaseTarget]
	def setStarbaseTarget(self, iX, iY):
		self.iXStarbaseTarget = iX
		self.iYStarbaseTarget = iY
		
	def getXStarbaseTarget(self):
		return self.iXStarbaseTarget
	def setXStarbaseTarget(self, iValue):
		self.iXStarbaseTarget = iValue

	def getYStarbaseTarget(self):
		return self.iYStarbaseTarget
	def setYStarbaseTarget(self, iValue):
		self.iYStarbaseTarget = iValue
		
	def getUnitIDStarbaseMaker(self):
		return self.iUnitIDStarbaseMaker
	def setUnitIDStarbaseMaker(self, iValue):
		printd("Setting Starbase Maker Unit ID to %d for player %d" %(iValue, self.getID()))
		self.iUnitIDStarbaseMaker = iValue
		
	def saveData(self):
		
		aData = []
		
		aData.append(self.iID)
		aData.append(self.iStarbaseTargetValue)
		aData.append(self.iXStarbaseTarget)
		aData.append(self.iYStarbaseTarget)
		aData.append(self.iUnitIDStarbaseMaker)
		
		return aData
		
	def loadData(self, aData):
		
		iIterator = 0
		
		self.setID(aData[iIterator])
		iIterator += 1
		self.setStarbaseTargetValue(aData[iIterator])
		iIterator += 1
		self.setXStarbaseTarget(aData[iIterator])
#		self.iXStarbaseTarget = aData[iIterator]
		iIterator += 1
		self.setYStarbaseTarget(aData[iIterator])
#		self.iYStarbaseTarget = aData[iIterator]
		iIterator += 1
		self.setUnitIDStarbaseMaker(aData[iIterator])